from glstring import flatten
import argparse


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("-i", "--infile",
                        required=True,
                        help="input file",
                        type=str)

    parser.add_argument("-o", "--outfile",
                        required=True,
                        help="output file",
                        type=str)
    args = parser.parse_args()

    infile = args.infile
    outfile = args.outfile
    print("reading from ", infile)
    print("writing to ", outfile)

    fin = open(infile, 'r')
    fout = open(outfile, 'w')

    with fin as lines:
        for line in lines:
            (id, gl) = line.rstrip().split('%')
            fgl = flatten(gl)
            fout.write('%'.join([id, fgl]) + '\n')
    fin.close()
    fout.close()


if __name__ == '__main__':
    """The following will be run if file is executed directly,
    but not if imported as a module"""
    main()



# glstring
# module for working with glstrings

import re


# convert genotype ambiguity into allele ambiguity

# TODO: need to handle "^" character

def flatten (gls):
    # if gls contains ^
    if re.search("\^", gls):
        # loop over all loci
        return "^".join(flatten_loc(g) for g in gls.split("^"))
    else:
        return flatten_loc(gls)
        

def flatten_loc (gls):
    # if gls contains |
    if re.search("\|", gls):
        # loop over all genos
        typ1 = dict()
        typ2 = dict()
        for geno in gls.split("|"):
            # split on +
            if not re.search("\+", geno):
                print("geno ", geno, " has no +")
            t1, t2 = geno.split("+")
            # add to hash1, hash2
            typ1[t1]=1
            typ2[t2]=1

        # join keys by /
        newt1 = "/".join(sorted(typ1.keys()))
        newt2 = "/".join(sorted(typ2.keys()))
        # join these by +
        newgeno = "+".join([newt1, newt2])
        return(newgeno)
    else: 
        return (gls)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

#
#    pyars pyARS.
#    Copyright (c) 2018 Be The Match operated by National Marrow Donor Program. All Rights Reserved.
#
#    This library is free software; you can redistribute it and/or modify it
#    under the terms of the GNU Lesser General Public License as published
#    by the Free Software Foundation; either version 3 of the License, or (at
#    your option) any later version.
#
#    This library is distributed in the hope that it will be useful, but WITHOUT
#    ANY WARRANTY; with out even the implied warranty of MERCHANTABILITY or
#    FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public
#    License for more details.
#
#    You should have received a copy of the GNU Lesser General Public License
#    along with this library;  if not, write to the Free Software Foundation,
#    Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA.
#
#    > http://www.fsf.org/licensing/licenses/lgpl.html
#    > http://www.opensource.org/licenses/lgpl-license.php
#

"""
test_pyars
----------------------------------

Tests for `pyars` module.
"""
import os
import sys
import json
import unittest

from pyard import ARD


class TestPyard(unittest.TestCase):

    def setUp(self):
        self.ard = ARD(verbose=True)
        self.data_dir = os.path.dirname(__file__)
        self.assertIsInstance(self.ard, ARD)
        expected_json = self.data_dir + "/expected.json"
        with open(expected_json) as json_data:
            self.expected = json.load(json_data)
        pass

    def test_000_nomac(self):
        self.ardnomac = ARD(download_mac=False)
        self.assertIsInstance(self.ardnomac, ARD)
        self.assertFalse(self.ardnomac.download_mac)
        self.assertTrue(len(self.ardnomac.mac.keys()) == 0)
        self.assertTrue(self.ardnomac.redux("A*01:01:01", 'G') == "A*01:01:01G")
        self.assertTrue(self.ardnomac.redux("A*01:01:01", 'lg') == "A*01:01g")
        self.assertTrue(self.ardnomac.redux("A*01:01:01", 'lgx') == "A*01:01")
        self.assertTrue(self.ardnomac.redux("HLA-A*01:01:01", 'G') == "HLA-A*01:01:01G")
        self.assertTrue(self.ardnomac.redux("HLA-A*01:01:01", 'lg') == "HLA-A*01:01g")
        self.assertTrue(self.ardnomac.redux("HLA-A*01:01:01", 'lgx') == "HLA-A*01:01")
        pass

    def test_001_dbversions(self):
        for db in ['3310', '3300', '3290', '3280']:
            self.arddb = ARD(dbversion=db, download_mac=False)
            self.assertIsInstance(self.arddb, ARD)
            self.assertFalse(self.arddb.download_mac)
            self.assertTrue(self.arddb.dbversion == db)
            self.assertTrue(self.arddb.redux("A*01:01:01", 'G') == "A*01:01:01G")
            self.assertTrue(self.arddb.redux("A*01:01:01", 'lg') == "A*01:01g")
            self.assertTrue(self.arddb.redux("A*01:01:01", 'lgx') == "A*01:01")
        pass

    def test_002_remove_invalid(self):
        self.assertTrue(self.ard.redux("A*01:01:01", 'G') == "A*01:01:01G")
        pass

    def test_003_mac(self):
        self.assertTrue(self.ard.redux_gl("A*01:AB", 'G') == "A*01:01:01G/A*01:02")
        self.assertTrue(self.ard.redux_gl("HLA-A*01:AB", 'G') == "HLA-A*01:01:01G/HLA-A*01:02")
        pass

    def test_004_redux_gl(self):
        for ex in self.expected['redux_gl']:
            glstring = ex['glstring']
            ard_type = ex['ard_type']
            expected_gl = ex['expected_gl']
            self.assertTrue(self.ard.redux_gl(glstring, ard_type) == expected_gl)
        pass

    def test_005_mac_G(self):
        self.assertTrue(self.ard.redux("A*01:01:01", 'G') == "A*01:01:01G")
        pass







# -*- coding: utf-8 -*-
#
#    seqann Sequence Annotation
#    Copyright (c) 2017 Be The Match operated by National Marrow Donor Program. All Rights Reserved.
#
#    This library is free software; you can redistribute it and/or modify it
#    under the terms of the GNU Lesser General Public License as published
#    by the Free Software Foundation; either version 3 of the License, or (at
#    your option) any later version.
#
#    This library is distributed in the hope that it will be useful, but WITHOUT
#    ANY WARRANTY; with out even the implied warranty of MERCHANTABILITY or
#    FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public
#    License for more details.
#
#    You should have received a copy of the GNU Lesser General Public License
#    along with this library;  if not, write to the Free Software Foundation,
#    Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA.
#
#    > http://www.fsf.org/licensing/licenses/lgpl.html
#    > http://www.opensource.org/licenses/lgpl-license.php
#

from pprint import pformat
from typing import TypeVar, Type
from six import iteritems
from .util import deserialize_model

T = TypeVar('T')


class Model(object):
    # dataTypes: The key is attribute name and the value is attribute type.
    data_types = {}

    # attributeMap: The key is attribute name and the value is json key in definition.
    attribute_map = {}

    @classmethod
    def from_dict(cls: Type[T], dikt) -> T:
        """
        Returns the dict as a model
        """
        return deserialize_model(dikt, cls)

    def to_dict(self):
        """
        Returns the model properties as a dict

        :rtype: dict
        """
        result = {}

        for attr, _ in iteritems(self.data_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value

        return result

    def to_str(self):
        """
        Returns the string representation of the model

        :rtype: str
        """
        return pformat(self.to_dict())

    def __repr__(self):
        """
        For `print` and `pprint`
        """
        return self.to_str()

    def __eq__(self, other):
        """
        Returns true if both objects are equal
        """
        return self.__dict__ == other.__dict__

    def __ne__(self, other):
        """
        Returns true if both objects are not equal
        """
        return not self == other


# -*- coding: utf-8 -*-

#
#    pyard pyARD
#    Copyright (c) 2018 Be The Match operated by National Marrow Donor Program. All Rights Reserved.
#
#    This library is free software; you can redistribute it and/or modify it
#    under the terms of the GNU Lesser General Public License as published
#    by the Free Software Foundation; either version 3 of the License, or (at
#    your option) any later version.
#
#    This library is distributed in the hope that it will be useful, but WITHOUT
#    ANY WARRANTY; with out even the implied warranty of MERCHANTABILITY or
#    FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public
#    License for more details.
#
#    You should have received a copy of the GNU Lesser General Public License
#    along with this library;  if not, write to the Free Software Foundation,
#    Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA.
#
#    > http://www.fsf.org/licensing/licenses/lgpl.html
#    > http://www.opensource.org/licenses/lgpl-license.php
#
import copy
import re
import urllib.request
import zipfile
from datetime import datetime, date

import pandas as pd
from six import integer_types, iteritems


def all_macs(csv_file, data_dir, url='https://hml.nmdp.org/mac/files/numer.v3.zip'):
    urllib.request.urlretrieve(url, 'numeric.v3.zip')
    zip_ref = zipfile.ZipFile('numeric.v3.zip', 'r')
    zip_ref.extractall(data_dir)
    zip_ref.close()
    data = []
    out_file = data_dir + "/numer.v3.txt"
    with open(out_file, 'r') as f:
        for line in f:
            line = line.rstrip()
            if re.search("^\D", line) and not re.search("CODE", line) and not re.search("LAST", line):
                data.append(line.split("\t"))
        f.close()
    df = pd.DataFrame(data, columns=['Code', 'Alleles'])
    df.to_csv(csv_file, header=True, index=False)
    df['Alleles'] = df['Alleles'].apply(lambda x: x.split("/"))
    mac_dict = df.set_index("Code").to_dict('index')
    return mac_dict


def pandas_explode(df, column_to_explode):
    """
    Similar to Hive's EXPLODE function, take a column with iterable elements, and flatten the iterable to one element 
    per observation in the output table
    :param df: A dataframe to explod
    :type df: pandas.DataFrame
    :param column_to_explode: 
    :type column_to_explode: str
    :return: An exploded data frame
    :rtype: pandas.DataFrame
    """
    # Create a list of new observations
    new_observations = list()
    # Iterate through existing observations
    for row in df.to_dict(orient='records'):
        # Take out the exploding iterable
        explode_values = row[column_to_explode]
        del row[column_to_explode]
        # Create a new observation for every entry in the exploding iterable & add all of the other columns
        for explode_value in explode_values:
            # Deep copy existing observation
            new_observation = copy.deepcopy(row)
            # Add one (newly flattened) value from exploding iterable
            new_observation[column_to_explode] = explode_value
            # Add to the list of new observations
            new_observations.append(new_observation)
    # Create a DataFrame
    return_df = pd.DataFrame(new_observations)
    # Return
    return return_df


def _deserialize(data, klass):
    """
    Deserializes dict, list, str into an object.

    :param data: dict, list or str.
    :param klass: class literal, or string of class name.

    :return: object.
    """
    if data is None:
        return None

    if klass in integer_types or klass in (float, str, bool):
        return _deserialize_primitive(data, klass)
    elif klass == object:
        return _deserialize_object(data)
    elif klass == date:
        return deserialize_date(data)
    elif klass == datetime:
        return deserialize_datetime(data)
    else:
        return deserialize_model(data, klass)


def _deserialize_primitive(data, klass):
    """
    Deserializes to primitive type.

    :param data: data to deserialize.
    :param klass: class literal.

    :return: int, long, float, str, bool.
    :rtype: int | long | float | str | bool
    """
    try:
        value = klass(data)
    except UnicodeEncodeError:
        value = unicode(data)
    except TypeError:
        value = data
    return value


def _deserialize_object(value):
    """
    Return a original value.

    :return: object.
    """
    return value


def deserialize_date(string):
    """
    Deserializes string to date.

    :param string: str.
    :type string: str
    :return: date.
    :rtype: date
    """
    try:
        from dateutil.parser import parse
        return parse(string).date()
    except ImportError:
        return string


def deserialize_datetime(string):
    """
    Deserializes string to datetime.

    The string should be in iso8601 datetime format.

    :param string: str.
    :type string: str
    :return: datetime.
    :rtype: datetime
    """
    try:
        from dateutil.parser import parse
        return parse(string)
    except ImportError:
        return string


def deserialize_model(data, klass):
    """
    Deserializes list or dict to model.

    :param data: dict, list.
    :type data: dict | list
    :param klass: class literal.
    :return: model object.
    """
    instance = klass()

    if not instance.swagger_types:
        return data

    for attr, attr_type in iteritems(instance.swagger_types):
        if data is not None \
                and instance.attribute_map[attr] in data \
                and isinstance(data, (list, dict)):
            value = data[instance.attribute_map[attr]]
            setattr(instance, attr, _deserialize(value, attr_type))

    return instance


def _deserialize_list(data, boxed_type):
    """
    Deserializes a list and its elements.

    :param data: list to deserialize.
    :type data: list
    :param boxed_type: class literal.

    :return: deserialized list.
    :rtype: list
    """
    return [_deserialize(sub_data, boxed_type)
            for sub_data in data]


def _deserialize_dict(data, boxed_type):
    """
    Deserializes a dict and its elements.

    :param data: dict to deserialize.
    :type data: dict
    :param boxed_type: class literal.

    :return: deserialized dict.
    :rtype: dict
    """
    return {k: _deserialize(v, boxed_type)
            for k, v in iteritems(data)}



# -*- coding: utf-8 -*-

#
#    pyard
#    Copyright (c) 2020 Be The Match operated by National Marrow Donor Program. All Rights Reserved.
#
#    This library is free software; you can redistribute it and/or modify it
#    under the terms of the GNU Lesser General Public License as published
#    by the Free Software Foundation; either version 3 of the License, or (at
#    your option) any later version.
#
#    This library is distributed in the hope that it will be useful, but WITHOUT
#    ANY WARRANTY; with out even the implied warranty of MERCHANTABILITY or
#    FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public
#    License for more details.
#
#    You should have received a copy of the GNU Lesser General Public License
#    along with this library;  if not, write to the Free Software Foundation,
#    Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA.
#
#    > http://www.fsf.org/licensing/licenses/lgpl.html
#    > http://www.opensource.org/licenses/lgpl-license.php
#
import re
import os
import pickle
import urllib.request
import pandas as pd
import functools
from .smart_sort import smart_sort_comparator
from .util import pandas_explode
from .util import all_macs
from operator import is_not
import functools
from functools import partial
from typing import Dict
import logging
import pathlib

ismac = lambda x: True if re.search(":\D+", x) else False


# a module shouldn't decide the logging config; thats up to the calling programo

#logging.basicConfig(format='%(asctime)s - %(name)s - %(levelname)s - %(message)s', datefmt='%m/%d/%Y %I:%M:%S %p', level=logging.INFO)

import string

alpha = list(string.ascii_lowercase) + list(string.ascii_uppercase)


# Have GFE ARD be in pyARD because requiring
# pygfe would be a big download.
def getvalue(a):
    return int("".join(list(a)[0:len(list(a))-1])) if list(a)[len(list(a))-1] in alpha else int(a)


def loci_sort(a, b):
    la = a.split(":")
    lb = b.split(":")
    a1, a2 = int(la[0].split("*")[1]), getvalue(la[1])
    b1, b2 = int(lb[0].split("*")[1]), getvalue(lb[1])
    if a1 > b1:
        return 1
    elif a1 == b1:
        if a2 > b2:
            return 1
        elif a2 == b2:
            return 0
        else:
            return -1
    else:
        return -1


class ARD(object):
    '''
    classdocs
    '''
    def __init__(self, dbversion: str='Latest',
                 download_mac: bool=True,
                 verbose: bool=False,
                 remove_invalid: bool=True,
                 data_dir: str=None):
        """
        ARD -
        :param dbversion:
        :type dbversion: str
        """
        self.data_types = {
            'dbversion': str,
            'verbose': bool,
            'download_mac': bool,
            'remove_invalid': bool,
            'G': Dict,
            'lg': Dict,
            'lgx': Dict
        }
        self.attribute_map = {
            'dbversion': 'dbversion',
            'G': 'G',
            'lg': 'lg',
            'lgx': 'lgx',
            'verbose': 'verbose',
            'download_mac': 'download_mac',
            'remove_invalid': 'remove_invalid'
        }

        self.mac = {}
        self._verbose = verbose
        self._dbversion = dbversion
        self._download_mac = download_mac
        self._remove_invalid = remove_invalid

        self.HLA_regex = re.compile("^HLA-")

        # TODO: add check for valid ARD type
        # TODO: add check for valid db version

        # List of expression characters
        expre_chars = ['N', 'Q', 'L', 'S']

        # Set data directory where all the downloaded files will go
        if data_dir is None:
            data_dir = os.path.dirname(__file__)
        else:
            pathlib.Path(data_dir).mkdir(exist_ok=True)

        imgt_hla_url = 'https://raw.githubusercontent.com/ANHIG/IMGTHLA/'
        ars_url = imgt_hla_url + dbversion + '/wmda/hla_nom_g.txt'
        allele_url = imgt_hla_url + dbversion + "/Allelelist.txt"

        ars_file = data_dir + '/hla_nom_g.' + str(dbversion) + ".txt"
        allele_file = data_dir + '/AlleleList.' + str(dbversion) + ".txt"
        mac_file = data_dir + "/mac.txt"
        mac_pickle = data_dir + "/mac.pickle"
        # dna_relshp.csv is part of the codebase
        broad_file = os.path.dirname(__file__) + "/dna_relshp.csv"

        # Downloading ARS file
        if not os.path.isfile(ars_file):
            if verbose:
                logging.info("Downloading " + str(dbversion) + " ARD file")
            urllib.request.urlretrieve(ars_url, ars_file)

        # Downloading allele list file
        if not os.path.isfile(allele_file):
            if verbose:
                logging.info("Downloading " + str(dbversion) + " allele list")
            urllib.request.urlretrieve(allele_url, allele_file)

        # Downloading MAC file
        if download_mac:
            if not os.path.isfile(mac_pickle):
                if verbose:
                    logging.info("Downloading MAC file")
                self.mac = all_macs(mac_file, data_dir=data_dir)

                # Writing dict to pickle file
                with open(mac_pickle, 'wb') as handle:
                    pickle.dump(self.mac, handle, protocol=pickle.HIGHEST_PROTOCOL)
            else:
                if verbose:
                    logging.info("Loading MAC file")
                with open(mac_pickle, 'rb') as handle:
                    self.mac = pickle.load(handle)

        with open(allele_file) as f:
            first_line = f.readline()
            f.close()

        sep = "," if re.search("#", first_line) else " "
        allele_data = []
        with open(allele_file, 'r') as f:
            for line in f:
                line = line.rstrip()
                if not re.search("#", line):
                    allele_data.append(line.split(sep))
            f.close()

        allele_df = pd.DataFrame(allele_data)
        if allele_df[:1].values.tolist()[0][1] == 'Allele':
            allele_df.columns = allele_df[:1].values.tolist()[0]
            idname = allele_df[:1].values.tolist()[0][0]
            allele_df.drop(0, inplace=True)
            allele_df = allele_df.rename(index=str, columns={idname: "ID"})
        else:
            allele_df.columns = ["ID", "Allele"]

        allele_df['2d'] = allele_df['Allele'].apply(lambda a:
                                     ":".join(a.split(":")[0:2]) +
                                     list(a)[-1] if list(a)[-1]
                                     in expre_chars and
                                     len(a.split(":")) > 2
                                     else ":".join(a.split(":")[0:2]))

        dfxx = pd.DataFrame(pd.Series(allele_df['2d'].unique().tolist()),
                                      columns=['Allele'])
        dfxx['1d'] = dfxx['Allele'].apply(lambda x: x.split(":")[0])
    
        # xxcodes maps a first field name to its expansion
        self.xxcodes = dfxx.groupby(['1d'])\
                           .apply(lambda x: list(x['Allele']))\
                           .to_dict()

        # defined broad XX codes
        dfbroad = pd.read_csv(broad_file, skiprows=1, dtype=str,
                         names=["Locus", "Broad", "Fam"], sep=",").dropna()

        dictbroad = dfbroad.groupby(['Locus','Broad']).apply(lambda x: list(x['Fam'])).to_dict()

        for (locus,broad) in dictbroad.keys():
            locusbroad="*".join([locus,broad])  
            for split in dictbroad[(locus,broad)]:
                locussplit="*".join([locus,split])
                if locusbroad in self.xxcodes:
                    self.xxcodes[locusbroad].extend(self.xxcodes[locussplit])
                else:
                    self.xxcodes[locusbroad] = self.xxcodes[locussplit]

        allele_df['3d'] = allele_df['Allele'].apply(lambda a:
                                 ":".join(a.split(":")[0:3]) +
                                 list(a)[-1] if list(a)[-1]
                                 in expre_chars and
                                 len(a.split(":")) > 3
                                 else ":".join(a.split(":")[0:3]))

        # all alleles are valid and also shortening to 3 and 2 fields
        self.valid = list(set(allele_df['Allele'].tolist()
                              + allele_df['2d'].tolist()
                              + allele_df['3d'].tolist()))
        # use a dict
        self.valid_dict={}
        for i in self.valid:
            self.valid_dict[i]=True

        # Loading ARS file into pandas
        # TODO: Make skip dynamic in case the files are not consistent
        df = pd.read_csv(ars_file, skiprows=6,
                         names=["Locus", "A", "G"], sep=";").dropna()

        df['Locus'] = df['Locus'].apply(lambda l: l.split("*")[0])
        df['A'] = df[['Locus', 'A']].apply(lambda row: [row['Locus'] + "*" + a
                                                        for a in
                                                        row['A'].split("/")
                                                        ],
                                           axis=1)
        df['G'] = df[['Locus', 'G']].apply(lambda row: "*".join([row['Locus'],
                                                                 row['G']]),
                                           axis=1)

        df = pandas_explode(df, 'A')

        df['2d'] = df['A'].apply(lambda a:
                                 ":".join(a.split(":")[0:2]) +
                                 list(a)[-1] if list(a)[-1]
                                 in expre_chars and
                                 len(a.split(":")) > 2
                                 else ":".join(a.split(":")[0:2]))

        df['3d'] = df['A'].apply(lambda a:
                                 ":".join(a.split(":")[0:3]) +
                                 list(a)[-1] if list(a)[-1]
                                 in expre_chars and
                                 len(a.split(":")) > 3
                                 else ":".join(a.split(":")[0:3]))

        df_values = df.drop_duplicates(['2d', 'G'])['2d']\
                      .value_counts().reset_index()\
                      .sort_values(by='2d', ascending=False)
        multiple_Glist = df_values[df_values['2d'] > 1]['index'].tolist()
        self.dup_g = df[df['2d'].isin(multiple_Glist)][['G', '2d']]\
                                .drop_duplicates()\
                                .groupby('2d', as_index=True).agg("/".join)\
                                .to_dict()['G']

        df['lg'] = df['G'].apply(lambda a:
                                 ":".join(a.split(":")[0:2]) + "g")

        df['lgx'] = df['G'].apply(lambda a:
                                  ":".join(a.split(":")[0:2]))

        # Creating dictionaries with allele->ARS group mapping
        self._G = pd.concat([df.drop(['A', 'lg', 'lgx', '3d'], axis=1)
                               .rename(index=str,
                                       columns={"2d": "A"})[['A', 'G']],
                            df.drop(['A', 'lg', 'lgx', '2d'], axis=1)
                              .rename(index=str,
                                      columns={"3d": "A"})[['A', 'G']],
                            df[['A', 'G']]],
                            ignore_index=True).set_index('A').to_dict()['G']

        self._lg = pd.concat([df.drop(['A', 'G', 'lgx', '3d'], axis=1)
                                .rename(index=str,
                                        columns={"2d": "A"})[['A', 'lg']],
                             df.drop(['A', 'G', 'lgx', '2d'], axis=1)
                               .rename(index=str,
                                       columns={"3d": "A"})[['A', 'lg']],
                             df[['A', 'lg']]],
                             ignore_index=True).set_index('A').to_dict()['lg']

        self._lgx = pd.concat([df.drop(['A', 'lg', 'G', '3d'], axis=1)
                                 .rename(index=str,
                                         columns={"2d": "A"})[['A', 'lgx']],
                              df.drop(['A', 'lg', 'G', '2d'], axis=1)
                                .rename(index=str,
                                        columns={"3d": "A"})[['A', 'lgx']],
                              df[['A', 'lgx']]],
                              ignore_index=True).set_index('A').to_dict()['lgx']

    @property
    def dbversion(self) -> str:
        """
        Gets the dbversion of this ARS.

        :return: The dbversion of this ARS.
        :rtype: str
        """
        return self._dbversion

    @property
    def verbose(self) -> bool:
        """
        Gets the verbose of this ARS.

        :return: The verbose of this ARS.
        :rtype: bool
        """
        return self._verbose

    @property
    def download_mac(self) -> bool:
        """
        Gets the download_mac of this ARS.

        :return: The download_mac of this ARS.
        :rtype: bool
        """
        return self._download_mac

    @property
    def remove_invalid(self) -> bool:
        """
        Gets the remove_invalid of this ARS.

        :return: The remove_invalid of this ARS.
        :rtype: bool
        """
        return self._remove_invalid


    @property
    def G(self):
        """
        Gets the G of this ARS.

        :return: The G of this ARS.
        :rtype: Dict
        """
        return self._G

    @property
    def lg(self):
        """
        Gets the lg of this ARS.

        :return: The lg of this ARS.
        :rtype: Dict
        """
        return self._lg

    @property
    def lgx(self):
        """
        Gets the lgx of this ARS.

        :return: The lgx of this ARS.
        :rtype: Dict
        """
        return self._lgx

    @functools.lru_cache(maxsize=None)
    def redux(self, allele: str, ars_type: str) -> str:
        """
        Does ARS reduction with allele and ARS type

        :param allele: An HLA allele.
        :type: str
        :param ars_type: The ARS ars_type.
        :type: str
        :return: ARS reduced allele
        :rtype: str
        """

        # PERFORMANCE: precompiled regex
        # dealing with leading HLA-

        if self.HLA_regex.search(allele):
            hla, allele_name = allele.split("-")
            return "-".join(["HLA", self.redux(allele_name, ars_type)])

        if ars_type == "G" and allele in self._G:
            if allele in self.dup_g:
                return self.dup_g[allele]
            else:
                return self.G[allele]
        elif ars_type == "lg" and allele in self._lg:
            return self.lg[allele]
        elif ars_type == "lgx" and allele in self._lgx:
            return self.lgx[allele]
        else:
            if self.remove_invalid:
                if allele in self.valid:
                    return allele
                else:
                    return
            else:
                return allele

    @functools.lru_cache(maxsize=None)
    def redux_gl(self, glstring: str, redux_type: str) -> str:
        """
        Does ARS reduction with allele and ARS type

        :param allele: An HLA allele.
        :type: str
        :param ars_type: The ARS ars_type.
        :type: str
        :return: ARS reduced allele
        :rtype: str
        """
        
        if not self.isvalid_gl(glstring):
            return ""

        if re.search("\^", glstring):
            return "^".join(sorted(set([self.redux_gl(a, redux_type) for a in glstring.split("^")]), key=functools.cmp_to_key(smart_sort_comparator)))

        if re.search("\|", glstring):
            return "|".join(sorted(set([self.redux_gl(a, redux_type) for a in glstring.split("|")]), key=functools.cmp_to_key(smart_sort_comparator)))

        if re.search("\+", glstring):
            return "+".join(sorted([self.redux_gl(a, redux_type) for a in glstring.split("+")], key=functools.cmp_to_key(smart_sort_comparator)))

        if re.search("\~", glstring):
            return "~".join([self.redux_gl(a, redux_type) for a in glstring.split("~")])

        if re.search("/", glstring):
            return "/".join(sorted(set([self.redux_gl(a, redux_type) for a in glstring.split("/")]), key=functools.cmp_to_key(smart_sort_comparator)))

        loc_allele = glstring.split(":")
        loc_name, code = loc_allele[0], loc_allele[1]
       
        # handle XX codes
        # test that they are valid
        if(ismac(glstring) and glstring.split(":")[1] == "XX") and loc_name in self.xxcodes:
            loc, n = loc_name.split("*")
            return self.redux_gl("/".join(sorted(self.xxcodes[loc_name], key=functools.cmp_to_key(smart_sort_comparator))), redux_type)

        if ismac(glstring) and code in self.mac:
            if re.search("HLA-", glstring):
                hla, allele_name = glstring.split("-")
                loc_name, code = allele_name.split(":")
                loc, n = loc_name.split("*")
                alleles = list(filter(lambda a: a in self.valid,
                                      [loc_name + ":" + a if len(a) <= 3
                                       else loc + "*" + a
                                       for a in self.mac[code]['Alleles']]))
                return self.redux_gl("/".join(sorted(["HLA-" + a for a in alleles], key=functools.cmp_to_key(smart_sort_comparator))), redux_type)
            else:
                loc, n = loc_name.split("*")
                alleles = list(filter(lambda a: a in self.valid,
                                      [loc_name + ":" + a if len(a) <= 3
                                       else loc + "*" + a
                                       for a in self.mac[code]['Alleles']]))
                return self.redux_gl("/".join(sorted(alleles, key=functools.cmp_to_key(smart_sort_comparator))), redux_type)
        return self.redux(glstring, redux_type)

    def isvalid(self, allele: str) -> bool:
        """
        Determines validity of an allele

        :param allele: An HLA allele.
        :type: str
        :return: allele or empty
        :rtype: bool
        """
        if not ismac(allele):
            # PERFORMANCE: use hash instead of allele in "list"
            # return allele in self.valid
            return self.valid_dict.get(allele, False)
        return True

    def isvalid_gl(self, glstring: str) -> bool:
        """
        Determine validity of glstring

        :param glstring
        :type: str
        :return: result
        :rtype: bool
        """
        
        if re.search("\^", glstring):
            return(all(list(map(self.isvalid_gl,glstring.split("^")))))
        if re.search("\|", glstring):
            return(all(list(map(self.isvalid_gl,glstring.split("|")))))
        if re.search("\+", glstring):
            return(all(list(map(self.isvalid_gl,glstring.split("+")))))
        if re.search("\~", glstring):
            return(all(list(map(self.isvalid_gl,glstring.split("~")))))
        if re.search("/", glstring):
            return(all(list(map(self.isvalid_gl,glstring.split("/")))))

        # what falls through here is an allele
        return(self.isvalid(glstring))

    def mac_toG(self, allele: str) -> str:
        """
        Does ARS reduction with allele and ARS type

        :param allele: An HLA allele.
        :type: str
        :param ars_type: The ARS ars_type.
        :type: str
        :return: ARS reduced allele
        :rtype: str
        """
        loc_name, code = allele.split(":")
        loc, n = loc_name.split("*")
        if code in self.mac:
            alleles = list(filter(lambda a: a in self.valid,
                                  [loc_name + ":" + a if len(a) <= 3
                                   else loc + "*" + a
                                   for a in self.mac[code]['Alleles']]))
            group = list(filter(partial(is_not, None),
                         set([self.toG(allele=a)
                              for a in alleles])))
            if "X" in group:
                return None
            else:
                return "/".join(group)

        else:
            return None

    def toG(self, allele: str) -> str:
        """
        Does ARS reduction to the G group level

        :param allele: An HLA allele.
        :type: str
        :return: ARS G reduced allele
        :rtype: str
        """
        if allele in self.G:
            if allele in self.dup_g:
                return self.dup_g[allele]
            else:
                return self.G[allele]
        else:
            return "X"

    def expand_mac(self, allele: str):
        """
        Exapnds mac codes

        :param allele: An HLA allele.
        :type: str
        :return: List
        :rtype: List
        """
        loc_name, code = allele.split(":")
        loc, n = loc_name.split("*")
        if len(loc.split("-")) == 2:
            loc = loc.split("-")[1]
            loc_name = loc_name.split("-")[1]

        if code in self.mac:
            return list(filter(lambda a: a in self.valid,
                               [loc_name + ":" + a if len(a) <= 3
                                else loc + "*" + a
                                for a in self.mac[code]['Alleles']]))
        else:
            return ''




import functools
import re

expr_regex = re.compile('[NQLSGg]')

@functools.lru_cache(maxsize=None)
def smart_sort_comparator(a1, a2):
    """
    Natural sort 2 given alleles.

    Python sorts strings lexographically but HLA alleles need
    to be sorted by numerical values in each field of the HLA nomenclature.

    :param a1: first allele
    :param a2: second allele
    """

    # Check to see if they are the same alleles
    if a1 == a2:
        return 0


    # remove any non-numerics
    a1 = re.sub(expr_regex, '', a1)
    a2 = re.sub(expr_regex, '', a2)
    # Extract and Compare first fields first
    a1_f1 = int(a1[a1.find('*')+1:a1.find(':')])
    a2_f1 = int(a2[a2.find('*')+1:a2.find(':')])

    if a1_f1 < a2_f1:
        return -1
    if a1_f1 > a2_f1:
        return 1

    # If the first fields are equal, try the 2nd fields
    a1_f2 = int(a1[a1.find(':')+1:])
    a2_f2 = int(a2[a2.find(':')+1:])

    if a1_f2 < a2_f2:
        return -1
    if a1_f2 > a2_f2:
        return 1

    # All fields are equal
    return 0

def smart_sort_alleles(a1, a2):
    """
    Natural sort 2 given alleles.

    Python sorts strings lexographically but HLA alleles need
    to be sorted by numerical values in each field of the HLA nomenclature.

    :param a1: first allele
    :param a2: second allele
    """
    # Check to see if they are the same alleles
    if a1 == a2:
        return [a1, a2]

    # Extract and Compare first fields first
    a1_f1 = int(a1[a1.find('*')+1:a1.find(':')])
    a2_f1 = int(a2[a2.find('*')+1:a2.find(':')])

    if a1_f1 < a2_f1:
        return [a1, a2]
    if a1_f1 > a2_f1:
        return [a2, a1]

    # If the first fields are equal, try the 2nd fields
    a1_f2 = int(a1[a1.find(':')+1:])
    a2_f2 = int(a2[a2.find(':')+1:])

    if a1_f2 < a2_f2:
        return [a1, a2]
    if a1_f2 > a2_f2:
        return [a2, a1]

    # All fields are equal
    return [a1, a2]
